pub mod raas {
    #[cfg(feature = "async")]
    use tokio::io::{AsyncReadExt, AsyncWriteExt};

    pub mod register {
        include!(concat!(env!("OUT_DIR"), "/raas.register.rs"));
    }

    pub mod bot {
        pub mod roll {
            include!(concat!(env!("OUT_DIR"), "/raas.bot.roll.rs"));
        }
    }

    pub mod ping {
        include!(concat!(env!("OUT_DIR"), "/raas.ping.rs"));
    }

    pub mod cmd {
        include!(concat!(env!("OUT_DIR"), "/raas.cmd.rs"));
    }

    pub mod resp {
        include!(concat!(env!("OUT_DIR"), "/raas.resp.rs"));
    }

    #[derive(Debug, Clone)]
    pub struct RaasMessage {
        pub len: u32,
        pub msg: Vec<u8>,
    }

    impl RaasMessage {
        pub fn new(msg: Vec<u8>) -> Self {
            Self {
                len: msg.len() as u32,
                msg,
            }
        }

        pub fn from_bytes(bytes: &[u8]) -> Self {
            let (len_bytes, msg) = bytes.split_at(std::mem::size_of::<u32>());
            let mut arr = [0u8; 4];
            arr.copy_from_slice(len_bytes);
            let len = u32::from_be_bytes(arr);

            Self {
                len,
                msg: msg.to_vec(),
            }
        }

        pub fn into_bytes(mut self) -> Vec<u8> {
            let mut bytes = Vec::new();

            bytes.extend_from_slice(&self.len.to_be_bytes());

            bytes.append(&mut self.msg);

            bytes
        }
    }

    #[cfg(not(feature = "async"))]
    pub fn recv_raas_msg<T: std::io::Read>(stream: &mut T) -> Result<RaasMessage, std::io::Error> {
        let mut message_len_bytes = [0u8; 4];
        stream.read_exact(&mut message_len_bytes)?;

        let len = u32::from_be_bytes(message_len_bytes);

        let mut message = vec![0u8; len as usize];

        stream.read_exact(&mut message)?;

        Ok(RaasMessage { len, msg: message })
    }

    #[cfg(not(feature = "async"))]
    pub fn send_raas_msg<T: std::io::Write>(
        stream: &mut T,
        data: Vec<u8>,
    ) -> Result<(), std::io::Error> {
        let msg = RaasMessage::new(data);

        stream.write_all(&msg.into_bytes())?;
        Ok(())
    }

    #[cfg(feature = "async")]
    pub async fn recv_raas_msg<T: AsyncReadExt + std::marker::Unpin>(
        stream: &mut T,
    ) -> Result<RaasMessage, std::io::Error> {
        let mut message_len_bytes = [0u8; 4];
        stream.read_exact(&mut message_len_bytes).await?;

        let len = u32::from_be_bytes(message_len_bytes);

        let mut message = vec![0u8; len as usize];

        stream.read_exact(&mut message).await?;

        Ok(RaasMessage { len, msg: message })
    }

    #[cfg(feature = "async")]
    pub async fn send_raas_msg<T: AsyncWriteExt + std::marker::Unpin>(
        stream: &mut T,
        data: Vec<u8>,
    ) -> Result<(), std::io::Error> {
        let msg = RaasMessage::new(data);

        stream.write_all(&msg.into_bytes()).await?;
        Ok(())
    }
}

#[cfg(test)]
mod test {
    use crate::raas;
    use crate::raas::{recv_raas_msg, send_raas_msg, RaasMessage};

    #[cfg(not(feature = "async"))]
    use std::io::BufReader;

    #[cfg(feature = "async")]
    use tokio::io::BufReader;

    #[test]
    fn create_cmd() {
        let cmd = raas::cmd::Command {
            id: 0,
            timestamp: 0,
            cmd: Some(raas::cmd::command::Cmd::RollCmd(raas::bot::roll::RollCmd {
                cmd: Some(raas::bot::roll::roll_cmd::Cmd::Roll(
                    raas::bot::roll::Roll { rotations: 1 },
                )),
            })),
        };

        assert_eq!(cmd.id, 0);
        assert_eq!(cmd.timestamp, 0);
        assert!(cmd.cmd.is_some());
    }

    #[test]
    fn encode_message() {
        let data = vec![0, 1, 2, 3, 4];
        let msg = RaasMessage::new(data.clone());

        let msg_bytes = msg.into_bytes();

        assert_eq!(msg_bytes.len(), data.len() + std::mem::size_of::<u32>());
    }

    #[test]
    fn decode_message() {
        let data: Vec<u8> = vec![0, 0, 0, 5, 0, 1, 2, 3, 4];
        let msg = RaasMessage::from_bytes(&data);

        assert_eq!(msg.len, 5);
        assert_eq!(msg.msg, data[4..]);
    }

    #[test]
    fn encode_decode_message() {
        let original_message = RaasMessage::new(vec![0x55, 0xff]);

        let data = original_message.clone().into_bytes();

        let decoded_message = RaasMessage::from_bytes(&data);

        assert_eq!(decoded_message.len, decoded_message.len);
        assert_eq!(decoded_message.msg, original_message.msg);
    }

    #[cfg(not(feature = "async"))]
    #[test]
    fn send_recv_test_raas_message() {
        let mut msg_buffer = Vec::new();
        let message = Vec::from(b"Hello World");

        send_raas_msg(&mut msg_buffer, message.clone()).unwrap();

        let mut reader = BufReader::new(msg_buffer.as_slice());
        let out_msg = recv_raas_msg(&mut reader).unwrap();

        assert_eq!(message.len() as u32, out_msg.len);
        assert_eq!(message, out_msg.msg);
    }

    #[cfg(feature = "async")]
    #[tokio::test]
    async fn send_recv_test_raas_message() {
        let mut msg_buffer = Vec::new();
        let message = Vec::from(b"Hello World");

        send_raas_msg(&mut msg_buffer, message.clone())
            .await
            .unwrap();

        let mut reader = BufReader::new(msg_buffer.as_slice());
        let out_msg = recv_raas_msg(&mut reader).await.unwrap();

        assert_eq!(message.len() as u32, out_msg.len);
        assert_eq!(message, out_msg.msg);
    }
}
